from __future__ import annotations
from .cursor import sqlalchemy
from .table_types import BulkInsert, BulkUpdate, BulkDelete
from sqlalchemy import Integer, Column
import json
from pathlib import Path


class Table:
    id = Column(Integer, primary_key=True, autoincrement=True)

    @classmethod
    def all(cls) -> list[Table]:
        with sqlalchemy.app_context():
            return sqlalchemy.db.session.query(cls).all()

    @classmethod
    def find_by(cls, **kwargs) -> Table | None:
        if len(kwargs) == 0:
            return
        args = [cls.__dict__[param] == arg for param, arg in kwargs.items()]
        with sqlalchemy.app_context():
            return sqlalchemy.db.session.query(cls).filter(*args).first()

    @classmethod
    def find_all_by(cls, **kwargs) -> list[Table]:
        if len(kwargs) == 0:
            return []
        args = [cls.__dict__[param] == arg for param, arg in kwargs.items()]
        with sqlalchemy.app_context():
            record = sqlalchemy.db.session.query(cls).filter(*args).all()
            return record

    @classmethod
    def columns(cls) -> list[str]:
        return cls.metadata.tables[cls.__tablename__].columns.keys()

    @classmethod
    def first(cls) -> Table:
        with sqlalchemy.app_context():
            return sqlalchemy.db.session.query(cls).first()

    @classmethod
    def length(cls) -> int:
        with sqlalchemy.app_context():
            return sqlalchemy.db.session.query(cls).count()

    @classmethod
    def bulk_insert(cls, values: list) -> BulkInsert:
        with sqlalchemy.app_context():
            inserted_records = []
            failed = []
            for kwargs in values:
                try:
                    record = cls(**kwargs)
                    sqlalchemy.db.session.add(record)
                    inserted_records.append(record)
                except Exception:
                    failed.append(failed)
            sqlalchemy.db.session.commit()
            inserted_record_ids = [r.id for r in inserted_records]
            inserted = sqlalchemy.db.session.query(cls).filter(cls.id.in_(inserted_record_ids)).all()
            return {
                "inserted": inserted,
                "inserted_count": len(inserted),
                "failed": failed,
                "failed_count": len(failed),
            }

    @classmethod
    def bulk_update(cls, values: list) -> BulkUpdate:
        with sqlalchemy.app_context():
            updated_ids = []
            failed = []
            for kwargs in values:
                try:
                    if 'id' not in kwargs:
                        failed.append(kwargs)
                        continue
                    record = cls.find_by(id=kwargs['id'])
                    if record is None:
                        failed.append(kwargs)
                        continue
                    for key, val in kwargs.items():
                        setattr(record, key, val)
                    updated_ids.append(record.id)
                    sqlalchemy.db.session.add(record)
                except Exception:
                    failed.append(kwargs)
            sqlalchemy.db.session.commit()
            updated = sqlalchemy.db.session.query(cls).filter(cls.id.in_(updated_ids)).all()
            return {
                "updated": updated,
                "updated_count": len(updated),
                "failed": failed,
                "failed_count": len(failed),
            }

    @classmethod
    def bulk_delete(cls, items: list[Table]) -> BulkDelete:
        with sqlalchemy.app_context():
            deleted = []
            failed = []
            for item in items:
                try:
                    data = dict(item)
                    sqlalchemy.db.session.delete(item)
                    deleted.append(data)
                except Exception:
                    failed.append(item)
            sqlalchemy.db.session.commit()
            return {
                'deleted': deleted,
                'deleted_count': len(deleted),
                'failed': failed,
                'failed_count': len(failed),
            }

    @classmethod
    def drop_table(cls) -> bool:
        if cls.table_exists():
            with sqlalchemy.app_context():
                try:
                    cls.__table__.drop(sqlalchemy.db.engine)
                except Exception:
                    return False
        return not cls.table_exists()

    @classmethod
    def create_table(cls) -> bool:
        if not cls.table_exists():
            with sqlalchemy.app_context():
                try:
                    cls.__table__.create(sqlalchemy.db.engine, checkfirst=True)
                    return True
                except Exception:
                    return False

    @classmethod
    def clear_table(cls) -> None:
        if cls.table_exists():
            cls.drop_table()
            cls.create_table()

    @classmethod
    def table_exists(cls) -> bool:
        return sqlalchemy.has_table(cls.__tablename__)

    @classmethod
    def _seed_table(cls, path: Path or str) -> None:
        if not cls.table_exists():
            cls.create_table()
        file = Path(path)
        if not file.exists():
            raise FileNotFoundError(f'Could not find: {str(file)}')
        with open(file, 'r') as r:
            values = json.load(r)
            assert isinstance(values, list)
            cls.bulk_insert(values)

    @classmethod
    def seed_table(cls) -> None:
        ...

    @classmethod
    def group_by(cls):
        # TODO: add a group by
        # need execution to consoladate because
        # not all bots in the same server
        # e.g. group_by server_id
        pass

    def create(self) -> Table:
        with sqlalchemy.app_context():
            sqlalchemy.db.session.add(self)
            sqlalchemy.db.session.commit()
            return self.find_by(id=self.id)

    def delete(self) -> bool:
        with sqlalchemy.app_context():
            try:
                sqlalchemy.db.session.delete(self)
                sqlalchemy.db.session.commit()
                return True
            except Exception:
                return False

    def keys(self) -> list:
        return self.columns()

    def values(self) -> list:
        return [getattr(self, key) for key in self.keys()]

    def items(self):
        for key in self.columns():
            yield key, getattr(self, key)

    def __getitem__(self, key):
        assert key in self.keys(), f"'{key}' must be of {self.columns()}"
        return getattr(self, key)

    def __call__(self, *args, **kwargs):
        difference = set(kwargs.keys()).difference(set(self.columns()))
        error_message = f"{difference} are not Columns of '{self.__tablename__}': {self.columns()}"
        assert len(difference) == 0, error_message
        with sqlalchemy.app_context():
            for key, val in kwargs.items():
                setattr(self, key, val)
            sqlalchemy.db.session.add(self)
            sqlalchemy.db.session.commit()
            return self.find_by(id=self.id)

    def __str__(self):
        table_name = self.__tablename__.title().replace("_", "")
        items = []
        for k, v in dict(self).items():
            items.append(f"\033[34m{k}\033[90m=\033[0m{repr(v)}\033[0m")
        args = ', '.join(items)
        return f'<\033[96m{table_name}\033[0m({args})>\033[0m'

    def __repr__(self):
        table_name = self.__tablename__.title().replace("_", "")
        items = []
        for k, v in dict(self).items():
            items.append(f"{k}={repr(v)}")
        args = ', '.join(items)
        return f'{table_name}({args})'



